package controlservices;

import communicationservices.*;
import dataservices.*;
import java.net.*;
import java.util.*;
import snmp.*;

/**
 * <p>Title: ProtocolHandler</p>
 * <p>Description: This class acts as the main handler for incoming UDP packets,
 * making descisions on how the SNMP Proxy should respond to different
 * incoming SNMP requests.  In general, one Protocol Handler is created for
 * every incoming UDP message.</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * <p>Company: Printronix</p>
 * @author Christopher M. King (kingc@uci.edu)
 * @version 1.2
 */

// TO DO: Rewrite function interfaces so that they use the SNMPMessageInfo wrapper
// rather then the PDU so that the SNMP implementation may be more seperated.

public class ProtocolHandler
    extends Thread
    implements ProxyConstants
{
  // DECLARATIONS
  private DatagramPacket udpPacket;   // the UDP packet that was recieved
  private UDPComManager udpManager;
  private ProxySecurityManager securityManager;
  private ProxySessionManager sessionTable;
  private ProxyLog proxyLog;

  private ProxyTimer timer;
  private PDUProcessor messageHandler;
  private SNMPPacketProcessor  snmpPacketProcessor;

  private ProxySession session;

  private SNMPMessageInfo requestInfo;

  private InetAddress sourceIP;
  private int sourcePort;

  /**
   * Runs the thread of the ProtocolHandler, the main controller for
   * handling and incoming UDP packet
   */
  public ProtocolHandler(DatagramPacket dP,
                         UDPComManager uCM,
                         ProxySecurityManager secM,
                         ProxySessionManager sesM,
                         ProxyLog log,
                         int timeout,
                         String transFile)
  {
    udpPacket       = dP;
    udpManager      = uCM;
    securityManager = secM;
    sessionTable    = sesM;
    proxyLog        = log;

    timer = new ProxyTimer(timeout, this);
    messageHandler = new PDUProcessor(securityManager, proxyLog, transFile);
    snmpPacketProcessor = new SNMPPacketProcessor();
  }

  /**
   * Runs the thread of the ProtocolHandler, the main controller for
   * handling and incoming UDP packet
   */
  public void run()
  {
    // Major protocol steps in capitol letters:

    // -- BEGIN THE RESPONSE TIMER:
    timer.start();

    // -- EXTRACT SNMP MSG: attempt to extract the SNMP information from the UDP
    // -- packet and log it
    requestInfo = snmpPacketProcessor.extractSNMP(udpPacket);
    if(requestInfo == null)         // if the UDP packet did not contain valid
    { timer.stop();                 // SNMP info
      proxyLog.logEvent(PACKET_ERROR);
      return;
    }
    // -- CHECK SNMP MSG AUTHENTICITY: check the the SNMP message by checking
    // -- its community name with the SecurityManager
    if(!securityManager.verifyCommunityName(requestInfo.getCommunity()))
    { timer.stop();
      proxyLog.logEvent(SECURITY_ERROR);
      proxyLog.logEvent("from ip: " +requestInfo.getIP());
      return;
    }
    // -- RETRIEVE/CREATE SESSION: get the info associated with the udp source
    // -- stored by the proxy or create an new entry (keyed by IP source)
    session = sessionTable.findProxySession(requestInfo.getIP());
    if(session == null)
      session = sessionTable.openProxySession(requestInfo.getIP());

    // -- PROCESS REQUEST: Do what is needed to get/set what is wanted
    SNMPMessageInfo snmpResponse = this.requestHandler(requestInfo);

    // -- PROCESS RESPONSE: Send response where it needs to go
    this.responseHandler(snmpResponse);
  }

  /**
   * Processes an SNMP request and sends it to the appropropriate message
   * handler to generate an SNMP response
   *
   * @param requestInfo contains the request to process
   * @return an SNMPMessageInfo containing the appropriate SNMP response
   * to the request made.
   */
  private SNMPMessageInfo requestHandler(SNMPMessageInfo requestInfo)
  {
    int snmpPDUType = requestInfo.getType();
    SNMPPDU snmpPDU = null;
    try
    {
      snmpPDU = requestInfo.getMessage().getPDU();
    }
    catch(SNMPBadValueException e)
    {
       System.out.println(e.toString());
    }

    if(snmpPDU == null)
      return null;

    String managedIP = session.getManagedIP();
    SNMPMessage responseMessage;
    switch(snmpPDUType)
    { /////////////
      //   SET   //
      /////////////
      case SNMPMessageInfo.SET:
        // special case, part of SNMPProxy protocol to set managed ip
        if(messageHandler.isManagedIPRequest(snmpPDU))
          snmpPDU = messageHandler.handleSetManagedIPRequest(session, snmpPDU);
        else if(messageHandler.isTableRequest(snmpPDU))
          snmpPDU = messageHandler.handleTableSetRequest(session, snmpPDU);
        //if(messageHandler.isCommunityRequest(snmpPDU))
        else
          snmpPDU = messageHandler.handleSetRequest(session, snmpPDU);
        break;

      /////////////  NOTE: Currently, a GetNext will get the first entry in
      // GETNEXT //  any table.
      /////////////
      case SNMPMessageInfo.GETNEXT:
        // -- MAP INTO A GET REQUEST: increment OIDs so that GETNEXT request
        // -- may be treated as a GET request
        snmpPDU = messageHandler.incrementOIDs(snmpPDU);
        responseMessage = new SNMPMessage(0, securityManager.getCommunityName(), snmpPDU);
        requestInfo = new SNMPMessageInfo(requestInfo.getIP(),requestInfo.getPort(),responseMessage);

        // Note: Lack of break is intentional

      /////////////
      //   GET   //
      /////////////
      case SNMPMessageInfo.GET:
        // special case, part of protocol
        if(messageHandler.isManagedIPRequest(snmpPDU))
        {
           snmpPDU = messageHandler.handleGetManagedIPRequest(session, snmpPDU);
           responseMessage = new SNMPMessage(0, securityManager.getCommunityName(), snmpPDU);
        }
        else
        { // -- CHECK THE RESPONSE QUE: This may be a request whose response was not
          // -- generated before the timeout (i.e. was responseded to with a wait msg)
          // -- but was stored in the buffer for quicker resposne next time.
          SNMPMessageInfo snmpResponseInfo = this.quedResponseHandler(requestInfo);
          if(snmpResponseInfo != null)
            return snmpResponseInfo;
          // -- CHECK THE REQUEST QUE: This may be a repeated request for which the
          // -- resonse is still not ready (i.e. the DMC user is impatient).  In this
          // -- case, rather then que additional requests, return a wait
          snmpResponseInfo = this.quedRequestHandler(requestInfo);
          if(snmpResponseInfo != null)
            return snmpResponseInfo;
          // -- CHECK IF TABLE REQUEST: If table request, que entire table
          // -- contents so that future table requests may be looked up in the que.
          if(messageHandler.isTableRequest(snmpPDU))
          {
            Vector v = messageHandler.handleTableGetRequest(session, snmpPDU);
            if(v != null)
            {
              for(int i=0; i<v.size(); i++)
              {
                String community = securityManager.getCommunityName();
                session.queSNMPResponse(new SNMPMessageInfo(requestInfo.getIP(), requestInfo.getPort(), new SNMPMessage(SNMP_V1, community, (SNMPPDU)v.elementAt(i))));
              }
            }
            SNMPMessageInfo temp = this.quedResponseHandler(requestInfo);
            if(temp != null)
            return temp;
          }

          // -- MAKE A NEW REQUEST: que the request in the cache and handle
          // -- the request.
          session.queSNMPRequest(requestInfo);
          snmpPDU = messageHandler.handleGetRequest(session, snmpPDU);

          if(snmpPDU == null) // request was invalid
              session.dequeSNMPRequest(requestInfo);
        }
        break;

      // Error handling
      case SNMPMessageInfo.ERROR:
        proxyLog.logEvent(TYPE_ERROR);
        break;

      default:
        proxyLog.logEvent(UNKOWN_MESSAGE);
        break;
      }
    responseMessage = new SNMPMessage(0, securityManager.getCommunityName(), snmpPDU);
    return new SNMPMessageInfo(requestInfo.getIP(),requestInfo.getPort(),responseMessage);
  }

  private SNMPMessageInfo quedResponseHandler(SNMPMessageInfo requestInfo)
  {
    SNMPMessageInfo quedResponseInfo = session.dequeSNMPResponse(requestInfo);
    if(quedResponseInfo != null) // if response was found in buffer
    {
      try
      { int requestID;
        SNMPSequence varBindList;
        SNMPPDU snmpPDU;
        SNMPMessage responseMessage;
        InetAddress requestIP = requestInfo.getIP();
        int requestPort = requestInfo.getPort();

        requestID = requestInfo.getRequestID();
        varBindList = quedResponseInfo.getMessage().getPDU().getVarBindList();

        snmpPDU = messageHandler.constructResponsePDU(requestID, varBindList, 0);
        String community = securityManager.getCommunityName();
        responseMessage = new SNMPMessage(SNMP_V1, community, snmpPDU);
        requestInfo = new SNMPMessageInfo(requestIP, requestPort, responseMessage);
        return requestInfo;
      }
      catch(SNMPBadValueException e)
      { System.out.println(e.toString());
        proxyLog.logEvent(e.toString());
      }
    }
    return null;
  }

  private SNMPMessageInfo quedRequestHandler(SNMPMessageInfo requestInfo)
  {
    if(session.containsSNMPRequest(requestInfo)) // if response was found in buffer
    {
      try
      { int requestID;
        SNMPSequence varBindList;
        SNMPPDU snmpPDU;
        SNMPMessage responseMessage;
        InetAddress requestIP = requestInfo.getIP();
        int requestPort = requestInfo.getPort();

        requestID = requestInfo.getRequestID();
        varBindList = requestInfo.getMessage().getPDU().getVarBindList();


        snmpPDU = messageHandler.packMsgIntoPDU(requestInfo.getMessage().getPDU(),"Please Wait...");
                        String community = securityManager.getCommunityName();
        responseMessage = new SNMPMessage(SNMP_V1, community, snmpPDU);
        requestInfo = new SNMPMessageInfo(requestIP, requestPort, responseMessage);
        return requestInfo;
      }
      catch(SNMPBadValueException e)
      { System.out.println(e.toString());
        proxyLog.logEvent(e.toString());
      }
    }
    return null;
  }

  private void responseHandler(SNMPMessageInfo responseInfo)
  {
    if(responseInfo == null)
    { //proxyLog.logEvent(RESPONSE_ERROR);
      return;
    }
    // if request has been processed in time to respond, send it
    if(!timer.isExpired())
    {
      timer.stop();
      if(requestInfo != null)
      {
        DatagramPacket response = snmpPacketProcessor.createUDPResponse(responseInfo);
        udpManager.send(response);
        session.dequeSNMPRequest(responseInfo);
      }
    }
    // if time ran out, store response for next time
    else
    { session.queSNMPResponse(responseInfo);
      session.dequeSNMPRequest(responseInfo);
    }
  }

  public void timeoutHandler()
  {
    if(requestInfo == null)
    { proxyLog.logEvent(RESPONSE_ERROR);
      return;
    }
      DatagramPacket wait;
      wait = snmpPacketProcessor.createMessageReponse(requestInfo, WAIT_MESSAGE);
      udpManager.send(wait);
  }
}
