package dataservices;

import communicationservices.*;
import dataservices.*;

import snmp.*; //use SNMPMessage, SNMPSequence, etc.
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * <p>Title: PDUProcessor</p>
 * <p>Description: Handles a LOT of lower level grunt work so that the protocol
 * manager can be somewhat abstract.  Also, limits the protocol managers
 * dependency on SNMP implementation.</p>
 * <p>Copyright: Copyright (c) 2002</p>
 * @author Susan XHu
 * @version 1.0
 */

// NOTE: Much of this code was written by Susan who was unavailible for adding
// additional comments.  Several parts of this code can be confusiing so you
// may wish to discuss issues here with her.

public class PDUProcessor
{
  TreeMap cmdMap;
  XMLCommandProcessor cmdProc;
  XMLResponseParser xmlParser;
  ProxySecurityManager securityManager;
  String translationFile;

  /**
  *	constructor
  */
  public PDUProcessor(ProxySecurityManager sM, ProxyLog proxyLog, String transFile)
  {
    translationFile = transFile;
    cmdMap = new TreeMap();
    cmdProc = new XMLCommandProcessor();
    xmlParser = new XMLResponseParser();
    processTranslationFile();
    securityManager = sM;
  }

  /**
  *	load translate.txt into a TreeMap, key is oid
  */
  void processTranslationFile()
  {
    String fileName = translationFile;
    BufferedReader in;

    try
    {
      in = new BufferedReader(new FileReader(fileName));
      String translationEntryLine;
      int i=0;
      while ((translationEntryLine = in.readLine()) != null)
      {
        if (translationEntryLine.length()!=0)
        { if (translationEntryLine.charAt(0) != 35)		//ignore lines starting with #
          { StringTokenizer st = new StringTokenizer (translationEntryLine, "\t");
            if (st.countTokens() < 6)
            { System.out.println("Error: Translation file formated incorrectly at line "+i+", " + " with value \"" + st.nextToken() + "\", # tokens is " +st.countTokens());
              System.exit(1);
            }
            String oid = st.nextToken();
            String access = st.nextToken();
            String type = st.nextToken();
            String command = st.nextToken();
            String attribute = st.nextToken();
            String value = st.nextToken();

            TranslationTableEntry transEntry = new TranslationTableEntry(oid, access, type, value, command, attribute);
            cmdMap.put(oid, transEntry);
          }
        }
        ++i;
      }
    }
    catch (Exception e)
    {
      System.out.print("Error: Failed to process file " + fileName);
      e.printStackTrace();
    }

  }

  public boolean isManagedIPRequest(SNMPPDU thePDU)
  {
    boolean setIP = false;
    SNMPSequence varList = thePDU.getVarBindList();
    if (varList.size()==1)
    {
      SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
      TranslationTableEntry entry = (TranslationTableEntry)cmdMap.get(oid.toString());
      if(entry == null) return false;
      if (entry.getCommand().trim().equals("SET_MANAGED_IP"))
      setIP = true;
    }
    return setIP;
  }

  public boolean isTableRequest(SNMPPDU thePDU)
  {
    boolean table = false;
    SNMPSequence varList = thePDU.getVarBindList();
    if (varList.size()==1)
    {
      SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
      String g = oid.toString();
      g = g.substring(0,g.lastIndexOf("."));
      g = g.substring(0,g.lastIndexOf("."));    //want to check that if the prefix of the oid is the one for table

      TranslationTableEntry entry = (TranslationTableEntry)cmdMap.get(g);
      if(entry == null)
      {
        entry = (TranslationTableEntry)cmdMap.get(oid.toString());
        if(entry == null) return false;
      }

      if(entry.getType().equals("TABLE"))
      {
        table = true;
      }
    }
    return table;
  }

  // checks if the oid contains an oid substring of type TABLE
  public boolean isTableRequest(String oid)
  {
    TranslationTableEntry entry = (TranslationTableEntry)cmdMap.get(oid);
    // is this a TABLE oid?
    if(entry != null)
        if(entry.getType().equals("TABLE"))
          return true;

    oid = oid.substring(0,oid.lastIndexOf("."));
    entry = (TranslationTableEntry)cmdMap.get(oid);
    // is this a TABLE_ENTRY oid?
    if(entry != null)
        if(entry.getType().equals("TABLE"))
          return true;

    oid = oid.substring(0,oid.lastIndexOf("."));
    entry = (TranslationTableEntry)cmdMap.get(oid);
    // is this a TABLE_ENTRY sub entry
    if(entry != null)
        if(entry.getType().equals("TABLE"))
          return true;

    return false;
  }

  public SNMPPDU handleSetManagedIPRequest(ProxySession session, SNMPPDU snmpPDU)
  {
    String managedIP = this.extractManagedIP(snmpPDU);
    session.setManagedIP(managedIP);

    int requestID = snmpPDU.getRequestID();
    SNMPSequence varList = snmpPDU.getVarBindList();
    SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
    SNMPObject value = (SNMPObject)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(1));
    return constructResponsePDU(requestID, constructVarList(oid, value), SNMPBERCodec.SNMP_ERR_NOERROR);
  }

  public SNMPPDU handleGetManagedIPRequest(ProxySession session, SNMPPDU snmpPDU)
  {
      return this.packMsgIntoPDU(snmpPDU, session.getManagedIP());
  }

  private String extractManagedIP(SNMPPDU thePDU)
  {
    SNMPSequence varList = thePDU.getVarBindList();
    SNMPObject deviceIP = (SNMPObject)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(1));
    return deviceIP.toString();
  }

  public SNMPPDU incrementOIDs(SNMPPDU snmpPDU)
  {
    SNMPSequence varList = snmpPDU.getVarBindList();
    SNMPSequence varPair = new SNMPSequence();

    for(int i=0; i<varList.size(); i++)
    {
      varPair = (SNMPSequence)varList.getSNMPObjectAt(i);
      try
      { SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(i)).getSNMPObjectAt(0));
        ((SNMPObjectIdentifier)(varPair.getSNMPObjectAt(0))).setValue(this.incrementOIDs(oid.toString()));
        oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(i)).getSNMPObjectAt(0));
      }
      catch(SNMPBadValueException e)
      {
        System.out.println("Error: " + e.toString());
      }
      catch(NoSuchElementException ex)  // This is a temp fix
      {
            try
            { return new SNMPPDU(snmpPDU.getPDUType(),snmpPDU.getRequestID(),snmpPDU.getErrorIndex(), snmpPDU.getErrorIndex(), snmpPDU.getVarBindList());
            }
            catch(SNMPBadValueException e)
            {
              return this.packMsgIntoPDU(snmpPDU, "Error: Failed to handle invalid OID");
            }
      }
    }
    return snmpPDU;
  }

  private String incrementOIDs(String oid) throws NoSuchElementException
  {
    String newOid;
    SortedMap m = cmdMap.tailMap(oid+"\0");
    newOid = (String)m.firstKey();

    if(this.isTableRequest(newOid))       // increment past table
      newOid = this.incrementOIDs(newOid);

    return newOid;
  }

  public SNMPPDU handleGetRequest(ProxySession session, SNMPPDU thePDU)
  {
    String managedIP = session.getManagedIP();

    SNMPSequence varList = thePDU.getVarBindList();
    int requestID = thePDU.getRequestID();

    if (varList.size()>0)
    {
      SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
      TranslationTableEntry entry = (TranslationTableEntry)cmdMap.get(oid.toString());

      if(entry == null)
      {
        try
        {
          return new SNMPPDU(thePDU.getPDUType(),thePDU.getRequestID(),thePDU.getErrorIndex(), thePDU.getErrorIndex(), thePDU.getVarBindList());
        }
        catch(SNMPBadValueException e)
        {
          return this.packMsgIntoPDU(thePDU, "Error: Failed to handle invalid OID");
        }
      }
      String cmd = entry.getCommand().trim();
      String valueType = entry.getType().trim();
      Vector attrs = entry.getAttributes();
      SNMPSequence responseVarList = new SNMPSequence();

      if (cmd.equals("na")) //return profile value in PDU directly
      {
        try
        {
          SNMPObject itemValue = convert2SNMPObject(valueType, entry.getProfileValue());
          SNMPVariablePair thePair = new SNMPVariablePair(oid, itemValue);

          responseVarList = new SNMPSequence();
          responseVarList.addSNMPObject(thePair);
        }
        catch (Exception e)
        {
          System.err.println("error caught in handleGetRequest ...");
          System.exit(1);
        }
        return constructResponsePDU(requestID, responseVarList, SNMPBERCodec.SNMP_ERR_NOERROR);
      }
      String xmlToSend;
      Vector vals = new Vector();

      vals.add(cmd);
      if (deviceSpecific(attrs))
        vals.add(managedIP);

      xmlToSend = cmdProc.formGetCommand(attrs,vals);

      //using POEMSClient methods here
      POMESClient theClient = session.getPOMESClient();
      theClient.sendCommand(xmlToSend);
      String response = theClient.receiveResponse();
      session.releasePOMESClient();

      xmlParser.XMLparse(response);
      if (xmlParser.getSuccess() == true)
      { //use oid from original request SNMPPDU and value XMLResponseParser turned
        String value = xmlParser.getValue();
        SNMPObject snmpValue = convert2SNMPObject(valueType, value);
        responseVarList = constructVarList(oid, snmpValue);
        return constructResponsePDU(requestID, responseVarList, SNMPBERCodec.SNMP_ERR_NOERROR);
      }
    }
    return constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_GENERR);
  }

  /**
  *	receives a Set-type SNMPMessage, turn it into XMLMessage, send it to POEMS
  *	get an XML response back from POEMS, parse it into an SNMPMessage and return it
  */
  public SNMPPDU handleSetRequest(ProxySession session, SNMPPDU thePDU)
  {
    POMESClient theClient = session.getPOMESClient();
    String managedIP = session.getManagedIP();
    SNMPSequence varList = thePDU.getVarBindList();
    int requestID = thePDU.getRequestID();

    if (varList.size()== 1)
    {
      SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
      SNMPObject value = (SNMPObject)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(1));

      TranslationTableEntry entry = (TranslationTableEntry)cmdMap.get(oid.toString());
      String cmd = entry.getCommand().trim();
      Vector attrs = entry.getAttributes();

      String xmlToSend = "";
      if (deviceSpecific(attrs))
      {
        Vector vals = new Vector();

        vals.add(cmd);
        if (deviceSpecific(attrs))
          vals.add(managedIP);

        xmlToSend = cmdProc.formGetCommand(attrs,vals);
      }

      try
      {
        theClient.sendCommand(xmlToSend);
        String response = theClient.receiveResponse();
        session.releasePOMESClient();

        //parse response
        session.releasePOMESClient();
        xmlParser.XMLparse(response);
      }
      // allows the proxy to recover from bugs in POMES I/O
      // w/o causing a session to lock up.
      catch(Exception e)
      {
        session.releasePOMESClient();
        System.out.println("Error: An I/O failure to POMES has occured: " + e.toString());
      }
      if (xmlParser.getSuccess() == true)
      {
        //use oid, value from original request SNMPPDU
        return constructResponsePDU(requestID, constructVarList(oid, value), SNMPBERCodec.SNMP_ERR_NOERROR);
      }
      else //return error with original varList
         return constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_GENERR);

    }
    else //multiple oids, device specific, only a Set command is applicable
    {
      //each oid should map to the same command name in translate.txt; they belong the same table entry
      int oidNum = varList.size();
      String[] types = new String[oidNum];
      String[] cmdsTest = new String[oidNum];
      String[] attributes = new String[oidNum];
      SNMPObject [] values = new SNMPObject[oidNum];

      for (int i=0; i<oidNum; i++)
      { SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(i)).getSNMPObjectAt(0));
        values[i] = (SNMPObject)(((SNMPSequence)varList.getSNMPObjectAt(i)).getSNMPObjectAt(1));
        TranslationTableEntry entry = (TranslationTableEntry)cmdMap.get(oid.toString());
        types[i] = entry.getType().trim();
        cmdsTest[i] = entry.getCommand().trim();
        attributes[i] = entry.getAttributeList().trim();
      }

      for (int i=1; i<oidNum; i++)
      { if (!cmdsTest[i].equals(cmdsTest[0]))
        {
          System.exit(1);
        }
      }
      //create something like <SNMPTrapEntry  entryid="2" ipaddress="1.2.3.4" mask="1"  />
      String tag = attributes[0].substring(0, attributes[0].indexOf('.')+1);
      String objString = "<" + tag + " ";
      boolean isDelete = false;
      //test: is it a Set- or Delete- Table-entry, look for second field specificly
      isDelete = (attributes[1]=="0.0.0.0") || (attributes[1] == "");
      for (int i=0; i<oidNum; i++)
      {
        objString = objString + attributes[i] + "=\"" + values[i].toString() + "\" ";
      }
      objString = objString + "  />";

      String xmlToSend = "";
      theClient.sendCommand(xmlToSend);
      String response = theClient.receiveResponse();
      session.releasePOMESClient();

      //parse response
      xmlParser.XMLparse(response);
      if (xmlParser.getSuccess() == true)
      { //use varList from original request SNMPPDU
        return constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_NOERROR);
      }
      else //turn error with original varList
        return constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_GENERR);
    }
  }

  public Vector handleTableGetRequest(ProxySession session, SNMPPDU thePDU)
  {
    POMESClient theClient = session.getPOMESClient();
    String managedIP = session.getManagedIP();

    SNMPSequence varList = thePDU.getVarBindList();
    int requestID = thePDU.getRequestID();
    String entryId = "0";

    if (varList.size()==1)
    {
      SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
      String asdf = oid.toString();
      String tryOid = asdf;
      tryOid = tryOid.substring(0,tryOid.lastIndexOf("."));
      entryId = tryOid.substring(tryOid.lastIndexOf(".") + 1); // get second to last number from OID
      tryOid = tryOid.substring(0,tryOid.lastIndexOf("."));
      if(isTableRequest(oid.toString()))
        asdf = tryOid;
      TranslationTableEntry entry;
      entry = (TranslationTableEntry)cmdMap.get(asdf);

      if((TranslationTableEntry)cmdMap.get(asdf) == null)
      {  // Maps table entry into first entry to retrieve data from the
         // the translation file
         String modOid = asdf;
         modOid = modOid.substring(0,tryOid.lastIndexOf("."));
         modOid = modOid.substring(0,tryOid.lastIndexOf("."));
         modOid += "1" + asdf.substring(asdf.lastIndexOf(".", asdf.length()));
         entry = (TranslationTableEntry)cmdMap.get(asdf);
      }
      if(entry == null)
        return new Vector();

      String cmd = entry.getCommand().trim();
      String valueType = entry.getType().trim();
      //String attribute = entry.getAttributeList().trim();
      Vector attrs = entry.getAttributes();
      SNMPSequence responseVarList = new SNMPSequence();

      String xmlToSend;
      Vector vals = new Vector();

      vals.add(cmd);
      if (deviceSpecific(attrs))
        vals.add(managedIP);

      xmlToSend = cmdProc.formGetCommand(attrs,vals);

        theClient.sendCommand(xmlToSend);

        String response = theClient.receiveResponse();
        session.releasePOMESClient();
        xmlParser.XMLparse(response);
        Vector tables = xmlParser.getTableAttrVals();

        Vector PDUs = new Vector();

        String val = null;
        String shortOID = asdf;

        if(tables.size() < 1)
        {
            for(int q = 0; q < 3; ++q)
            {
                entry = (TranslationTableEntry)cmdMap.get(shortOID + ".1" + "." + Integer.toString(q+1));
                val = entry.getType();
                String newOid = shortOID + ".1." + Integer.toString(q+1);
                SNMPObjectIdentifier updatedOid = null;
                try{ updatedOid = new SNMPObjectIdentifier(newOid); }
                catch(Exception e){ e.toString();}

                SNMPObject snmpValue = convert2SNMPObject(val, "(none)");
                responseVarList = constructVarList(updatedOid, snmpValue);
                PDUs.add(constructResponsePDU(requestID, responseVarList, SNMPBERCodec.SNMP_ERR_NOERROR));
            }
            return PDUs;
        }

            // get the entry id
            // get the table element from vector
            // edit
            // call set command

        for(int k = 0; k < tables.size(); ++k)
        {
           for(int j = 0; j < ((Vector)tables.elementAt(k)).size(); ++j)
           {
              entry = (TranslationTableEntry)cmdMap.get(shortOID + ".1" + "." + Integer.toString(j+1));
              val = entry.getType();
              String newOid = shortOID + "." + Integer.toString(k+1) + "." + Integer.toString(j+1);
              SNMPObjectIdentifier updatedOid = null;

              try{ updatedOid = new SNMPObjectIdentifier(newOid); }

              catch(Exception e){ System.out.println("Error: " + e.toString());}

              String value = (String)((Vector)tables.elementAt(k)).elementAt(j);
              SNMPObject snmpValue = convert2SNMPObject(val, value);
              responseVarList = constructVarList(updatedOid, snmpValue);
              PDUs.add(constructResponsePDU(requestID, responseVarList, SNMPBERCodec.SNMP_ERR_NOERROR));
            }
        }
        Integer entryNum = new Integer(entryId);
        int i = entryNum.intValue();
        // If the request was for a valid (existing) entry, return the table contents as a vector of PDUs
        if( i > 0 && i <= tables.size() + 1) // +1 for the empty slot
        {  return PDUs;}
    }
    //else // make this default case
    {
      SNMPPDU errorPdu = constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_GENERR);
      Vector t = new Vector();
      t.add(errorPdu);
      return t;
    }
  }

	//helper function
  public SNMPObject convert2SNMPObject (String valueTypeString, String value)
  {
    if(valueTypeString.equals("STRING"))
      valueTypeString = "snmp.SNMPOctetString";
    else if(valueTypeString.equals("INTEGER"))
      valueTypeString = "snmp.SNMPInteger";
    else if(valueTypeString.equals("IPADDRESS"))
      valueTypeString = "snmp.SNMPIPAddress";
    else if(valueTypeString.equals("TABLE"))
      valueTypeString = "snmp.SNMPOctetString";
    else if(valueTypeString.equals("TABLE_ENTRY"))
      valueTypeString = "snmp.SNMPOctetString";
    else if(valueTypeString.equals("HEX"))
    {  value = convertHex2Binary(value);     // Fix for data type conflict
       valueTypeString = "snmp.SNMPInteger"; // between POMES and Printronix MIB
    }
    else valueTypeString = "snmp.SNMPUnknownObject";

    SNMPObject itemValue = null;
    try
    {
      Class valueClass = Class.forName(valueTypeString);
      itemValue = (SNMPObject)valueClass.newInstance();
      itemValue.setValue(value);
    }
    catch (Exception e)
    {
      System.out.println("");
    }
    return itemValue;
  }

  //helper function
  boolean deviceSpecific(Vector attrList)
  {
    //if device specific, the first word in attriubteList has to be "device_ip"
    if (attrList.size() >=0) // > length of "na"
      return(attrList.contains("device"));
    else
      return false;
  }

  public SNMPPDU constructResponsePDU( int requestID, SNMPSequence responseVarList, int error)
  {
    int errorStatus = 0;
    int errorIndex = 0;

    SNMPPDU responsePDU = null;
    try
    {
      responsePDU = new SNMPPDU(SNMPBERCodec.SNMPGETRESPONSE,
                                requestID,
                                errorStatus,
                                errorIndex,
                                responseVarList);
    }
    catch (Exception e)
    {
      System.err.println();
      e.printStackTrace();
      System.exit(1);
    }
    return responsePDU;
  }

  public SNMPPDU packMsgIntoPDU(SNMPPDU snmpPDU, String message)
  {
    String community = securityManager.getCommunityName().toString();

    try
    { SNMPOctetString testName = new SNMPOctetString(message);    //testing string

      int errorStatus = 0;
      int errorIndex = 0;

      SNMPSequence varList = snmpPDU.getVarBindList();
      SNMPSequence responseVarList = new SNMPSequence();

      for (int i=0; i<varList.size(); i++)
      { //parse oid, supply the value for that oid
        SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(i)).getSNMPObjectAt(0));
        SNMPVariablePair nextPair = new SNMPVariablePair(oid, testName);
        responseVarList.addSNMPObject(nextPair);
      }

      int requestID = snmpPDU.getRequestID();

      return new SNMPPDU(SNMPBERCodec.SNMPGETRESPONSE, requestID, errorStatus, errorIndex, responseVarList);
    }
    catch(Exception e){}

    return null;
  }

  SNMPSequence constructVarList(SNMPObjectIdentifier oid, SNMPObject value )
  {
    SNMPSequence varList = new SNMPSequence();;
    try
    {
      SNMPVariablePair thePair = new SNMPVariablePair(oid, value);
      varList.addSNMPObject(thePair);
    }
    catch(Exception e)
    {
      System.err.println("Error: Encoutered a problem while contstructing a var list.");
    }
    return varList;
  }

  public SNMPPDU handleTableSetRequest(ProxySession session, SNMPPDU thePDU)
  {
    POMESClient theClient = session.getPOMESClient();
    String managedIP = session.getManagedIP();

    SNMPSequence varList = thePDU.getVarBindList();
    int requestID = thePDU.getRequestID();

    if(varList.size() < 1)
      return null;

      SNMPObjectIdentifier oid = (SNMPObjectIdentifier)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(0));
      String oidStr = oid.toString();
      String tableOid = oid.toString();
      tableOid = tableOid.substring(0,tableOid.lastIndexOf("."));
      tableOid = tableOid.substring(0,tableOid.lastIndexOf("."));
      String entryRootOid = tableOid + ".1";

      TranslationTableEntry entry;
      entry = (TranslationTableEntry)cmdMap.get(tableOid);
      if((TranslationTableEntry)cmdMap.get(oid.toString()) == null)
      {      // patern match to translation file
             String modOid = oid.toString();
             modOid = modOid.substring(0,tableOid.lastIndexOf("."));
             modOid = modOid.substring(0,tableOid.lastIndexOf("."));
             modOid += "1" + oidStr.substring(oidStr.lastIndexOf(".", oidStr.length()));
             entry = (TranslationTableEntry)cmdMap.get(modOid);
      }
      if(entry == null)
        return constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_GENERR);

      TranslationTableEntry tableEntryRoot = (TranslationTableEntry)cmdMap.get(entryRootOid);

      String cmd = entry.getCommand().trim();
      String valueType = entry.getType().trim();
      //String attribute = entry.getAttributeList().trim();
      SNMPSequence responseVarList = new SNMPSequence();
      Vector cmdAttrNames = entry.getAttributes();

      String xmlToSend = "";
      Vector cmdAttrVals = new Vector();
      cmdAttrVals.add(cmd);
      if(deviceSpecific(cmdAttrNames))
        cmdAttrVals.add(managedIP);

      // -- GET THE TABLE
      xmlToSend = cmdProc.formGetCommand(cmdAttrNames,cmdAttrVals);
      theClient.sendCommand(xmlToSend);
      String response = theClient.receiveResponse();
      xmlParser.XMLparse(response);

      Vector tableAttrVals = xmlParser.getTableAttrVals();
      Vector tableAttrNames = xmlParser.getTableAttrNames();

      String index2 = oid.toString().substring(oid.toString().lastIndexOf(".") + 1, oid.toString().length() );
      String ff = oid.toString().substring(0 , oid.toString().lastIndexOf("."));

      String index = ff.substring(ff.lastIndexOf(".") + 1, ff.length() );

      int a = Integer.parseInt(index2);
      int b = Integer.parseInt(index);

       Vector entAttrVals = new Vector();
       Vector entAttrNames = new Vector();
      if(tableAttrVals.size() > b) // quick fix
        entAttrVals = (Vector)tableAttrVals.elementAt(b);
      if(tableAttrNames.size()> b)
      entAttrNames = (Vector)tableAttrNames.elementAt(b);
      Vector temp = new Vector();
      for(int i=0; i<entAttrVals.size(); i++)
      {
        if(i == b) // If this is the table entry element we want to modify
          temp.add(((SNMPObject)(((SNMPSequence)varList.getSNMPObjectAt(0)).getSNMPObjectAt(1))).toString());
        else
          temp.add((String)entAttrVals.elementAt(i));
      }
      entAttrVals = temp;
      String entName = xmlParser.getResponseObjectTag();

      // quick fix, retrieves the command from the translation file to prevent
      // the SetGet prefix bug.
      cmdAttrVals.setElementAt(tableEntryRoot.getCommand(),0);

      xmlToSend = cmdProc.formTableSetCommand(cmdAttrNames, cmdAttrVals, entName, entAttrNames, entAttrVals);

      theClient.sendCommand(xmlToSend);
      response = theClient.receiveResponse();
      session.releasePOMESClient();

     return constructResponsePDU(requestID, varList, SNMPBERCodec.SNMP_ERR_NOERROR);
  }

  private String convertHex2Binary(String hex)
  {
    String binary = "";
    for(int i = 0; i < 4; ++i)
    {
      switch(hex.charAt(i))
      {
        case('0'): binary += "0000"; break;
        case('1'): binary += "0001"; break;
        case('2'): binary += "0010"; break;
        case('3'): binary += "0011"; break;
        case('4'): binary += "0100"; break;
        case('5'): binary += "0101"; break;
        case('6'): binary += "0110"; break;
        case('7'): binary += "0111"; break;
        case('8'): binary += "1000"; break;
        case('9'): binary += "1001"; break;
        case('A'): binary += "1010"; break;
        case('B'): binary += "1011"; break;
        case('C'): binary += "1100"; break;
        case('D'): binary += "1101"; break;
        case('E'): binary += "1110"; break;
        case('F'): binary += "1111"; break;
      }
    }
    return binary;
  }
}